
#include "qobjects.h"

#include <cstdlib>
#include <ctime>

int main(int argc, char* argv[]){
   
//---------------------------------------initialize parameters

   const int seed = 23456;              // random seed 

   const double hbar = 1,                // fundamental constants 
                e  = 2.7182818,
                pi = 3.1415927;
		
   const std::complex<double> ei(0,1);
   
   const unsigned int N = 20,             // number of fock states
                      ns = 24000000,      // number of time steps 
                      ns_noise = ns,   
                      nt = 1;             // number of trajectories
   const int          n_out = 500;
   
   const double T = 120,                   // total time
                dt = T/ns,
		
                omega = 80,
                coupling_1 = 3,            // interaction strength 
                coupling_2 = 0.1,
				coupling_3 = 0.003*pi, 
                driving = 0,
                g = 0,                     // thermal noise rate
                xi = 2,                    // temperature parameter 
                k = 4;                     // measurement rate


//--------------------------------initialise files

   std::ofstream series("emeas.dat",std::ios_base::trunc);
      
//--------------------------------initialize operators 
                    
   qoperator I("identity",N), sI("identity",2),
             a("destroy",N),  sa("destroy",2),   a_1 = tensor(a,sI), a_2 = tensor(I,sa),
             adag("create",N), sadag("create",2), adag_1 = tensor(adag,sI), adag_2 = tensor(I,sadag),
             E_1 = adag_1*a_1, E_2 = adag_2*a_2, E = adag*a, x = a + adag,
             x_1 = a_1 + adag_1, x_2 = a_2 + adag_2,
             x2_1 = x_1*x_1, x2_2 = x_2*x_2;
             
// Hamiltonian    
   std::vector<qoperator> iHdt(3),iH2dt2(3);
   //iHdt[0] = (-ei*coupling_1*dt)*(a_1*adag_2 + adag_1*a_2);
   //iHdt[1] = (-ei*coupling_2*dt)*(a_1*adag_2 + adag_1*a_2);
   //iHdt[2] = (-ei*coupling_3*dt)*(a_1*adag_2 + adag_1*a_2);
   
   iHdt[0] = (-ei*omega*dt)*adag_1*a_1 + (-ei*omega*dt)*adag_2*a_2 + (-ei*coupling_1*dt)*x_1*x_2;
   iHdt[1] = (-ei*omega*dt)*adag_1*a_1 + (-ei*omega*dt)*adag_2*a_2 + (-ei*coupling_2*dt)*x_1*x_2;
   iHdt[2] = (-ei*omega*dt)*adag_1*a_1 + (-ei*omega*dt)*adag_2*a_2 + (-ei*coupling_3*dt)*x_1*x_2;
   
   for (int j=0;j<3;j++) iH2dt2[j] = 0.5*iHdt[j]*iHdt[j];
   
   //qoperator iHdt = (-ei*omega*dt)*adag_1*a_1 
   //               + (-ei*omega*dt)*adag_2*a_2 + (-ei*coupling*dt)*(a_1 + adag_1)*(a_2 + adag_2);
   //qoperator iHdt = (-ei*omega*dt)*adag*a; 

// observable to measure
   qoperator M_op = E_2;
   qoperator M2_op = M_op*M_op;
   
// noise operators
   qoperator B_op = x_2;
   qoperator B2_op = B_op*B_op;
   qoperator A_op = E_1, Ad_op = a_1*adag_1, C_op = a_1, Cd_op = adag_1;
   
// other operators   
   qoperator O_op = E_1, O2_op = O_op*O_op;
                
//--------------------------------initial state 

   //std::vector< std::complex<double> > psi_vec1(N); psi_vec1[2] = 1;             
   qstate psi_1("coherent",std::sqrt(5.0),N);
   qstate psi_2("up",2);
   qstate psi = tensor(psi_1,psi_2);   
   
//---------------------------------parameters for Wiener noise (dW's)

   std::vector<double> noise(2);     
   srand(seed); 
   double r8kdt = std::sqrt(8*k*dt);
   
//------------------------------------------------------------
//                        Simulation
//------------------------------------------------------------

   double EM_op = M_op.expect(psi).real(),  // initialise variables
          EO_op = O_op.expect(psi).real(),
          Ex = B_op.expect(psi).real(), 
          EO2 = O2_op.expect(psi).real(),
          EM2 = M2_op.expect(psi).real(),
          VO = EO2 - EO_op*EO_op,
          VM = EM2 - EM_op*EM_op,
          d_rec, d_rec_a, d_rec_adag; 
          
   time_t time0, *time_ptr = 0;       // timing initialisation
   time0 = time(time_ptr);
   
   // do multiple trajectories
   for (unsigned int j=0;j<nt;j++) {               
   
     // initial output 
     series << 0 << " " << Ex << " " << EO_op << " " << EM_op << " " 
			<< VO << " " << VM << " " << EM2 << std::endl; 
      
     // integrate a single trajectory
     for (unsigned int i=0;i<ns;i++) {         
          
       if (std::div(i,40000).rem==0) {std::cout << "T = " << i*dt << std::endl;}  // report progress
   
       // Hamiltonian Evolution (2nd order in dt)
       if (i*dt<=90) psi += iHdt[0]*psi + iH2dt2[0]*psi; 
       if (i*dt>90)  psi += iHdt[1]*psi + iH2dt2[1]*psi;
       if (i*dt>200)  psi += iHdt[2]*psi + iH2dt2[2]*psi;
       
       make_noise(noise); 
       for (int ni=0;ni<noise.size();ni++) { noise[ni] /= r8kdt; }  

       // Measurement Evolution (half order in dt)
       EM_op = M_op.expect(psi).real();
       d_rec = 4*k*dt*(EM_op + noise[0]);
       psi  += (-k*dt)*(M2_op*psi) + d_rec*(M_op*psi); 
       
       //std::cout << d_rec << std::endl;

       // x noise (no damping)
       //Ex = B_op.expect(psi).real();
       //d_rec = 4*g*dt*(Ex + noise2[i+j*ns_noise]); 
       //psi  += (-g*dt)*(B2_op*psi) + d_rec*(B_op*psi); 
       
       // quantum optical noise
//       d_rec_a = 4*g*(xi+1)*dt*(Ex/2 + noise2[i+j*ns_noise]);
//       d_rec_adag = 4*g*(xi)*dt*(Ex/2 + noise3[i+j*ns_noise]); 
//       psi  += (-g*(xi+1)*dt)*(A_op*psi) + d_rec_a*(C_op*psi) 
//               + (-g*(xi)*dt)*(Ad_op*psi) + d_rec_adag*(Cd_op*psi); 

       psi.normalise(); 
       
       //psi.display();
       
       // some expectation values
       EO_op = O_op.expect(psi).real();
       EO2 = O2_op.expect(psi).real();
       EM2 = M2_op.expect(psi).real();
       VO = EO2 - EO_op*EO_op;
       VM = EM2 - EM_op*EM_op;
       
       // output time series
       if (std::div(i,n_out).rem==0) {
         series << i*dt << " " << Ex << " " <<  EO_op << " " << EM_op << " " 
				<< VO << " " << VM << " " << EM2 << std::endl; 
       }

     }
     
   }
         
   //-----------------------
   // output simulation time
   //-----------------------

   std::cout << std::endl << "Simulation took " << difftime(time(time_ptr),time0) 
                          << " seconds" << std::endl;
                          
   series.close();
   	                 
   return 0;
}
